import os
import typing
import ntpath
import pathlib
import posixpath
from importlib import import_module
from typing import Any, ClassVar, Iterator, Optional, Type, TypeVar, Union, Callable, List, TYPE_CHECKING
from lazy.libz import Lib

from . import types
from . import base

_P = TypeVar('_P')


URI_PREFIXES = ('gs://', 's3://', 'minio://', 's3compat://')
_URI_SCHEMES = frozenset(('gs', 's3', 'minio', 's3compat'))
_URI_MAP_ROOT = {
    'gs://': '/gs/',
    's3://': '/s3/',
    'minio://': '/minio/',
    's3compat://': '/s3compat/'
}
_PROVIDER_MAP = {
    'gs': 'GoogleCloudStorage',
    's3': 'AmazonS3',
    'minio': 'MinIO',
    's3compat': 'S3Compatible'
}


PathLike = types.PathLike
ReadOnlyPath = base.ReadOnlyPath
ReadWritePath = base.ReadWritePath
PathLikeCls = Union[Type[ReadOnlyPath], Type[ReadWritePath]]

class _IOPath(pathlib.PurePath, base.ReadWritePath):
    """Pathlib-like API around `fsspec` providing Async Capabilities"""
    _PATH: ClassVar[types.ModuleType]
    _FSX: ClassVar[types.ModuleType] = None
    _FSX_LIB: str = None
    _FSX_MODULE: Optional[str] = None
    _FSX_CLS: Optional[str] = None
    _SYNC_FS: ClassVar[types.ModuleType] = None
    _ASYNC_FS: ClassVar[types.ModuleType] = None

    @classmethod
    def _ensure_lib(cls, *args, **kwargs):
        if cls._FSX is not None: return
        cls._FSX = Lib.import_lib(cls._FSX_LIB)
        if cls._FSX_MODULE: cls._FSX = import_module(cls._FSX_MODULE, package=cls._FSX_LIB)
        #cls._FSX = Lib.import_lib(cls._FSX_LIB)

    @classmethod
    def get_filesystem(cls, is_async: bool = False, *args, **kwargs):
        cls._ensure_lib()
        if is_async and cls._ASYNC_FS: return cls._ASYNC_FS
        if cls._SYNC_FS: return cls._SYNC_FS
        authz = cls.get_configz(*args, **kwargs)
        if is_async:
            cls._ASYNC_FS = getattr(cls._FSX, cls._FSX_CLS)(asynchronous = True, **authz)
            return cls._ASYNC_FS
        cls._SYNC_FS = getattr(cls._FSX, cls._FSX_CLS)(**authz)
        return cls._SYNC_FS
    
    @classmethod
    def get_configz(cls, *args, **kwargs):        
        return {}

    @property
    def async_fs(self) :
        if not self._ASYNC_FS:
            self.get_filesystem(is_async=True)
        return self._ASYNC_FS
    
    @property
    def sync_fs(self):
        if not self._SYNC_FS:
            self.get_filesystem()
        return self._SYNC_FS

    def __new__(cls: Type[_P], *parts: types.PathLike) -> _P:
        full_path = '/'.join(os.fspath(p) for p in parts)
        if not full_path.startswith(URI_PREFIXES): return super().__new__(cls, *parts)
        prefix = full_path[:5]
        new_prefix = _URI_MAP_ROOT[prefix]
        return super().__new__(cls, full_path.replace(prefix, new_prefix, 1))

    def _new(self: _P, *parts: types.PathLike) -> _P:
        """Create a new `Path` child of same type."""
        return type(self)(*parts)

    @property
    def _uri_scheme(self) -> Optional[str]:
        if (len(self.parts) >= 2 and self.parts[0] == '/' and self.parts[1] in _URI_SCHEMES): return self.parts[1]
        else: return None
    
    def get_bucket(self, prefix: bool = True) -> Optional[str]:
        """
        Returns the root bucket with optional prefix
        """
        if not self.is_cloud: return None
        parts = self.split_path()
        uri_scheme = self._uri_scheme
        if prefix: return f'{uri_scheme}://' + parts[0]
        return parts[0]
    
    def get_bucket_path(self, prefix: bool = True) -> Optional[str]:
        """
        Returns the root bucket + path with optional prefix
        removes the versioning
        """
        if not self.is_cloud: return None
        parts = self.split_path()
        uri_scheme = self._uri_scheme
        p = '/'.join(parts[:2])
        if prefix: return f'{uri_scheme}://' + p
        return p

    @property
    def bucket(self) -> Optional[str]:
        if self._uri_scheme: return self._PATH.join(f'{self._uri_scheme}://', self.parts[3])
        return None

    @property
    def bucket_path(self) -> Optional[str]:
        if self._uri_scheme: return self._PATH.join(*self.parts[3:])
        return None
    

    @property
    def _path_str(self) -> str:
        """
        Returns the `__fspath__` string representation.
        """
        uri_scheme = self._uri_scheme
        if uri_scheme: return self._PATH.join(f'{uri_scheme}://', *self.parts[2:])
        else: return self._PATH.join(*self.parts) if self.parts else '.'
    
    @property
    def string(self) -> str:
        """
        Returns the extension for a file
        """
        return self._path_str

    @property
    def _filename_str(self) -> str:
        """
        Returns the filename if is file, else ''
        """
        if self.is_file(): return self.parts[-1]
        return ''
    

    @property
    def extension(self) -> str:
        """
        Returns the extension for a file
        """
        return self.suffix

    @property
    def _cpath_str(self) -> str:
        """
        Returns the `__fspath__` string representation without the uri_scheme
        """
        uri_scheme = self._uri_scheme
        if uri_scheme: return self._PATH.join(*self.parts[2:])
        else: return self._PATH.join(*self.parts) if self.parts else '.'

    def __fspath__(self) -> str:
        return self._path_str

    def __str__(self) -> str:    # pylint: disable=invalid-str-returned
        return self._path_str

    def __repr__(self) -> str:
        return f'{type(self).__name__}({self._path_str!r})'

    def expanduser(self: _P) -> _P:
        """
        Returns a new path with expanded `~` and `~user` constructs.
        """
        return self._new(self._PATH.expanduser(self._path_str))
    

    def resolve(self: _P, strict: bool = False) -> _P:
        """
        Returns the abolute path.
        """
        if self.is_cloud: return self._new(self.as_posix())
        return self._new(self._PATH.abspath(self._path_str))
        
    def copydir(self: _P, dst: base.PathLike, ignore=['.git'], overwrite: bool = False, dryrun: bool = False):
        """
        Copies the Current Top Level Parent Dir to the Dst Dir without recursion
        """
        dst = self._new(dst)
        assert dst.is_dir(), 'Destination is not a valid directory'
        if not dryrun: dst.ensure_dir()
        copied_files = []
        fnames = self.listdir(ignore=ignore)
        curdir = self.absolute_parent
        for fname in fnames:
            dest_path = dst.joinpath(fname.relative_to(curdir))
            if not dryrun: fname.copy(dest_path, overwrite=overwrite, skip_errors=True)
            copied_files.append(dest_path)
        return copied_files


    def copydirs(self: _P, dst: base.PathLike, mode: str = 'shallow', pattern='*', ignore=['.git'], overwrite: bool = False, levels: int = 2, dryrun: bool = False):
        """Copies the Current Parent Dir to the Dst Dir.
        modes = [shallow for top level recursive. recursive for all nested]
        levels = number of recursive levels
        dryrun = returns all files that would have been copied without copying
        """
        assert mode in {'shallow', 'recursive'}, 'Invalid Mode Option: [shallow, recursive]'
        dst = self._new(dst)
        assert dst.is_dir(), 'Destination is not a valid directory'
        levels = max(1, levels)
        dst.ensure_dir()
        curdir = self.absolute_parent
        copied_files = []
        if levels > 1 and mode == 'recursive' and '/' not in pattern:
            for _ in range(1, levels): pattern += '/*'
        if self.is_dir() and not pattern.startswith('/'): pattern = '*/' + pattern
        fiter = curdir.glob(pattern) if mode == 'shallow' else curdir.rglob(pattern)
        fnames = [f for f in fiter if not bool(set(f.parts).intersection(ignore))]
        for f in fnames:
            dest_path = dst.joinpath(f.relative_to(curdir))
            if not dryrun:
                if f.is_dir(): dest_path.ensure_dir()
                else: f.copy(dest_path, overwrite=overwrite, skip_errors=True)
            copied_files.append(dest_path)
        return copied_files

    def listdir(self: _P, ignore=['.git'], skip_dirs=True, skip_files=False):
        fnames = [f for f in self.iterdir() if not bool(set(f.parts).intersection(ignore))]
        fnames = [f.resolve() for f in fnames]
        if skip_dirs:
            return [f for f in fnames if f.is_file()]
        if skip_files:
            return [f for f in fnames if f.is_dir()]
        return fnames

    def listdirs(self: _P, mode: str = 'shallow', pattern='*', ignore=['.git'], skip_dirs=True, skip_files=False, levels: int = 2):
        """Lists all files in current parent dir
        modes = [shallow for top level recursive. recursive for all nested]
        """
        assert mode in {'shallow', 'recursive'}, 'Invalid Mode Option: [shallow, recursive]'
        curdir = self.absolute_parent
        levels = max(1, levels)
        if levels > 1 and mode == 'recursive' and '/' not in pattern:
            for _ in range(1, levels): pattern += '/*'
        if self.is_dir() and not pattern.startswith('*/'): pattern = '*/' + pattern
        fiter = curdir.glob(pattern) if mode == 'shallow' else curdir.rglob(pattern)
        fnames = [f for f in fiter if not bool(set(f.parts).intersection(ignore))]
        if skip_dirs: return [f for f in fnames if f.is_file()]
        if skip_files: return [f for f in fnames if f.is_dir()]
        return list(fnames)

    def ensure_dir(self: _P, mode: int = 0o777, parents: bool = True, exist_ok: bool = True):
        """Ensures the parent directory exists, creates if not"""
        return self.absolute_parent.mkdir(mode=mode, parents=parents, exist_ok=exist_ok)
    
    def _get_dest(self: _P, dest: base.PathLike, recursive: bool = False, overwrite: bool = False, skip_errors: bool = False):
        """
        Validates the Destination
        """
        _dest = self._new(dest)
        if self.is_dir() and recursive: return _dest
        if _dest.is_dir(): _dest = _dest.joinpath(self._filename_str)
        if _dest.exists() and not overwrite:
            if skip_errors: return _dest
            raise ValueError(f'{_dest.as_posix()} exists and overwrite = False')
        return _dest

    @property
    def absolute_parent(self) -> _P:
        uri_scheme = self._uri_scheme
        if uri_scheme: return self._new(self._PATH.join(f'{uri_scheme}://', '/'.join(self.parts[2:-1])))
        p = self.resolve()
        if p.is_dir(): return p
        return p.parent

    @property
    def is_cloud(self) -> bool:
        return bool(self._uri_scheme)
    
    @property
    def cloud_provider(self) -> Optional[str]:
        return _PROVIDER_MAP.get(self._uri_scheme, None)

    @property
    def provider(self) -> Optional[str]:
        return _PROVIDER_MAP.get(self._uri_scheme, 'Local')

    @property
    def bucket(self) -> Optional[str]:
        raise NotImplementedError

    @property
    def bucket_path(self) -> Optional[str]:
        raise NotImplementedError

    @property
    def is_cloud(self) -> bool:
        return bool(self._uri_scheme)
    
    @property
    def is_gs(self) -> bool:
        return bool(self._uri_scheme == 'gs')
    
    @property
    def is_s3(self) -> bool:
        return bool(self._uri_scheme == 's3')
    
    @property
    def is_minio(self) -> bool:
        return bool(self._uri_scheme == 'minio')
    
    def exists(self) -> bool:
        """
        Returns True if self exists.
        """
        raise NotImplementedError

    def is_dir(self) -> bool:
        """
        Returns True if self is a directory.
        """
        raise NotImplementedError

    def is_file(self) -> bool:
        """
        Returns True if self is a file.
        """
        raise NotImplementedError

    def cat(self: _P, recursive: bool = False, on_error: Optional[str] = None, **kwargs):
        """
        Fetch (potentially multiple) paths contents
        """
        raise NotImplementedError
    
    def cat_file(self: _P, start: Optional[int] = None, end: Optional[int] = None, **kwargs):
        """
        Get the content of a file
        """
        raise NotImplementedError

    def copy(self: _P, dest: base.PathLike, overwrite: bool = False, skip_errors: bool = False) -> _P:
        """
        Copies the File to the Dir/File.
        """
        raise NotImplementedError
    
    def download(self: _P, dest: base.PathLike, recursive: bool = False, callback: Optional[Callable] = None, **kwargs):
        """
        Copy file(s) to local.
        """
        raise NotImplementedError
    
    def du(self: _P, total: Optional[int] = None, maxdepth: Optional[int] = None, **kwargs):
        """
        Fetch (potentially multiple) paths contents
        """
        raise NotImplementedError
    
    def find(self: _P, maxdepth: Optional[int] = None, withdirs: bool = False, detail: bool = False, prefix: Optional[str] = None, **kwargs):
        """
        List all files below path.
        """
        raise NotImplementedError
    
    def get_checksum(self: _P, refresh: bool = False, **kwargs):
        """
        Unique value for current version of file
        """
        raise NotImplementedError
    
    def get_file(self: _P, dest: base.PathLike, callback: Optional[Callable] = None, **kwargs):
        """
        Copy single remote file to local
        """
        raise NotImplementedError

    def get_files(self: _P, dest: base.PathLike, recursive: bool = False, callback: Optional[Callable] = None, **kwargs):
        """
        Copy file(s) to local.
        """
        raise NotImplementedError

    def glob(self: _P, pattern: str) -> Iterator[_P]:
        """
        Yielding all matching files (of any kind).
        """
        raise NotImplementedError

    def head(self: _P, size: Optional[int] = None):
        """ 
        Get the first size bytes from file
        """
        raise NotImplementedError

    def iterdir(self: _P) -> Iterator[_P]:
        """
        Iterates over the directory.
        """
        raise NotImplementedError


    def get_modified(self: _P, version_id: str = None, refresh: bool = False, **kwargs):
        """
        Return the last modified timestamp of file at path as a datetime
        """
        raise NotImplementedError
    
    def ls(self: _P, detail: bool = False, **kwargs):
        """
        List objects at path.
        """
        raise NotImplementedError
    
    def mkdir(self: _P, mode: int = 0o777, parents: bool = False, exist_ok: bool = False) -> None:
        """
        Create a new directory at this given path.
        """
        raise NotImplementedError

    def move(self: _P, dest: base.PathLike, recursive: bool = False, maxdepth: Optional[int] = None, **kwargs):
        """
        Move file(s) from one location to another
        """
        raise NotImplementedError
    
    def mv(self: _P, dest: base.PathLike, recursive: bool = False, maxdepth: Optional[int] = None, **kwargs):
        """
        Move file(s) from one location to another
        """
        raise NotImplementedError

    def open(self: _P, mode: str = 'r', encoding: Optional[str] = base.DEFAULT_ENCODING, errors: Optional[str] = None, **kwargs: Any) -> typing.IO[Union[str, bytes]]:
        """Opens the file."""
        raise NotImplementedError
    
    def put_file(self: _P, dest: base.PathLike, callback: Optional[Callable] = None, **kwargs):
        """
        Copy file from local.
        """
        raise NotImplementedError
    
    def put_files(self: _P, dest: base.PathLike, recursive: bool = False, callback: Optional[Callable] = None, **kwargs):
        """
        Copy file(s) from local.
        """
        raise NotImplementedError
    

    def remove(self: _P, recursive: bool = False, maxdepth: Optional[int] = None, **kwargs):
        """
        Delete files.
        """
        raise NotImplementedError

    def rename(self: _P, dest: base.PathLike, recursive: bool = False, maxdepth: Optional[int] = None, **kwargs):
        """
        Rename file or directory to the given target.
        """
        raise NotImplementedError    

    def replace(self: _P, target: base.PathLike) -> _P:
        """
        Replace file or directory to the given target.
        """
        raise NotImplementedError

    def rm(self: _P, recursive: bool = False, maxdepth: Optional[int] = None, **kwargs):
        """
        Delete files.
        """
        raise NotImplementedError
    
    def rm_file(self: _P, missing_ok: bool = True, **kwargs):
        """
        Delete a file
        """
        raise NotImplementedError

    def rmdir(self: _P, force: bool = False) -> None:
        """
        Remove the empty directory.
        """
        raise NotImplementedError
        
    def rmtree(self: _P) -> None:
        """
        Remove the directory.
        """
        raise NotImplementedError

    
    def sign(self: _P, expiration: int = 100, **kwargs):
        """
        Create a signed URL representing the given path
        Some implementations allow temporary URLs to be generated, as a way of delegating credentials
        """
        raise NotImplementedError
    
    def split_path(self: _P, **kwargs) -> List[str]:
        """
        Normalise path string into bucket and key
        """
        raise NotImplementedError
    
    def tail(self: _P, size: Optional[int] = None):
        """ 
        Get the last size bytes from file
        """
        raise NotImplementedError

    def touch(self: _P, truncate: bool = True, data = None, **kwargs):
        """
        Create empty file or truncate
        """
        raise NotImplementedError
    
    def unlink(self: _P, missing_ok: bool = True) -> None:
        """
        Remove this file or symbolic link.
        """
        raise NotImplementedError
    

    def url(self: _P, expires: int = 3600, client_method: str = 'get_object', **kwargs):
        """
        Generate presigned URL to access path by HTTP
        """
        raise NotImplementedError

    def upload(self: _P, dest: base.PathLike, recursive: bool = False, callback: Optional[Callable] = None, **kwargs):
        """
        Copy file(s) from local.
        """
        raise NotImplementedError

    @property
    def checksum(self):
        """
        Unique value for current version of file without refreshing
        """
        return self.get_checksum()
    
    @property
    def created(self):
        """
        Return the created timestamp of a file as a datetime
        """
        return self.get_checksum()
    

    @property
    def exist(self) -> bool:
        """
        Returns True if self exists.
        """
        return self.exists()

    @property
    def info(self):
        """
        Give details of entry at path
        """
        raise NotImplementedError
    
    @property
    def isdir(self) -> bool:
        """
        Returns True if self is a directory.
        """
        return self.is_dir()

    @property
    def isfile(self) -> bool:
        """
        Returns True if self is a file.
        """
        return self.is_file()

    @property
    def modified(self):
        """
        Return the last modified timestamp of file at path as a datetime without refreshing
        """
        return self.get_modified()
    
    @property
    def size(self: _P):
        """
        Size in bytes of file
        """
        raise NotImplementedError
    
    @property
    def stat(self: _P):
        """
        Give details of entry at path
        """
        raise NotImplementedError
    
    @property
    def home(self):
        """
        Returns the home directory
        """
        if self.is_cloud: return self._new(self.get_bucket(True))
        p = os.path.expanduser('~')
        return self._new(p)
    
    @property
    def userdir(self):
        return self.home



import upath
from fsspec.asyn import AsyncFileSystem
from fsspec.implementations.local import LocalFileSystem

class AsyncFSx(AsyncFileSystem, LocalFileSystem):
    pass

class PosixFSxPath(_IOPath, pathlib.PurePosixPath):
    """
    Pathlib-like API around `fsspec` providing Async Capabilities
    """
    _PATH = posixpath
    _FSX: 'LocalFileSystem' = None
    _SYNC_FS: 'LocalFileSystem' = None
    _ASYNC_FS: 'LocalFileSystem' = None
    _FSX_LIB: str = 'fsspec'
    _FSX_MODULE: Optional[str] = 'fsspec.implementations.local'
    _FSX_CLS: str = 'LocalFileSystem'

    @classmethod
    def _get_filesystem(cls, is_async: bool = False, *args, **kwargs):
        cls._ensure_lib()
        if is_async and cls._ASYNC_FS: return cls._ASYNC_FS
        if cls._SYNC_FS: return cls._SYNC_FS
        # remove this since we will have it be for sure available.
        #from fsspec.implementations.local import LocalFileSystem
        if is_async:
            cls._ASYNC_FS = AsyncFSx(asynchronous = True)
            return cls._ASYNC_FS
        cls._SYNC_FS = LocalFileSystem()
        return cls._SYNC_FS

    @property
    def async_fs(self) -> 'AsyncFSx':
        if not self._ASYNC_FS: self._get_filesystem(is_async=True)
        return self._ASYNC_FS
    
    @property
    def sync_fs(self) -> 'LocalFileSystem':
        if not self._SYNC_FS: self._get_filesystem()
        return self._SYNC_FS

    
    @classmethod
    def get_configz(cls, *args, **kwargs):
        return {}
    
    def exists(self) -> bool:
        """
        Returns True if self exists.
        """
        return self.sync_fs.exists(self._cpath_str)
    
    def expand_path(self, recursive: bool = False, maxdepth: Optional[int] = None):
        """
        Turn one or more globs or directories into a list of all matching paths to files or directories.
        """
        paths = self.sync_fs.expand_path(recursive = recursive, maxdepth = maxdepth)
        return [self._new(p) for p in paths]

    def is_dir(self) -> bool:
        """
        Returns True if self is a directory.
        """
        return self.sync_fs.isdir(self._cpath_str)

    def is_file(self) -> bool:
        """Returns True if self is a file."""
        return self.sync_fs.isfile(self._cpath_str)


    def cat(self, recursive: bool = False, on_error: Optional[str] = None, **kwargs):
        """
        Fetch (potentially multiple) paths contents
        """
        return self.sync_fs.cat(self._cpath_str, recursive, on_error, **kwargs)
    
    def cat_file(self, start: Optional[int] = None, end: Optional[int] = None, **kwargs):
        """
        Get the content of a file
        """
        return self.sync_fs.cat_file(self._cpath_str, start=start, end=end, **kwargs)

    def copy(self, dest: base.PathLike, recursive: bool = False, overwrite: bool = False, skip_errors: bool = False) -> _P:
        """
        Copies the File to the Dir/File.
        """
        _dest = self._get_dest(dest, recursive, overwrite, skip_errors)
        self.sync_fs.copy(self._cpath_str, _dest._cpath_str, recursive = recursive)
        return _dest

    
    def download(self, dest: base.PathLike, recursive: bool = False, callback: Optional[Callable] = None, overwrite: bool = False, skip_errors: bool = False, **kwargs):
        """
        Copy file(s) to local.
        """
        _dest = self._get_dest(dest, recursive, overwrite, skip_errors)
        self.sync_fs.download(self._cpath_str, _dest._cpath_str, recursive = recursive, callback = callback)
        return _dest
    
    def du(self, total: Optional[int] = None, maxdepth: Optional[int] = None, **kwargs):
        """
        Fetch (potentially multiple) paths contents
        """
        return self.sync_fs.du(self._cpath_str, total, maxdepth, **kwargs)
    
    def find(self, path: Optional[str] = None, maxdepth: Optional[int] = None, withdirs: bool = False, detail: bool = False, prefix: Optional[str] = None, **kwargs):
        """
        List all files below path.
        """
        p = self if not path else self.joinpath(path)
        return self.sync_fs.find(p, maxdepth=maxdepth, withdirs=withdirs, detail=detail, prefix=prefix, **kwargs)
    
    def get_checksum(self, refresh: bool = False, **kwargs):
        """
        Unique value for current version of file
        """
        return self.sync_fs.checksum(self._cpath_str, refresh=refresh, **kwargs)
    
    def get_file(self, dest: base.PathLike, callback: Optional[Callable] = None, overwrite: bool = False, skip_errors: bool = False, **kwargs):
        """
        Copy single remote file to local
        """
        _dest = self._get_dest(dest, overwrite=overwrite, skip_errors=skip_errors)
        return self.sync_fs.get_file(self._cpath_str, _dest._cpath_str, callback=callback, **kwargs)


    def get_files(self, dest: base.PathLike, recursive: bool = False, callback: Optional[Callable] = None, overwrite: bool = False, skip_errors: bool = False, **kwargs):
        """
        Copy file(s) to local.
        """
        _dest = self._get_dest(dest, recursive=recursive, overwrite=overwrite, skip_errors=skip_errors)
        return self.sync_fs.get(self._cpath_str, _dest._cpath_str, recursive=recursive, callback=callback, **kwargs)
        

    def glob(self, pattern: str, **kwargs) -> Iterator[_P]:
        """
        Yielding all matching files (of any kind).
        """
        uri_scheme = self._uri_scheme
        for f in self.sync_fs.glob(self._PATH.join(self._path_str, pattern)._cpath_str,  **kwargs):
            if self.is_cloud: yield self._new(f'{uri_scheme}://' + f)
            else: yield self._new(f)

    def head(self, size: Optional[int] = None):
        """ 
        Get the first size bytes from file
        """
        return self.sync_fs.head(self._cpath_str, size)

    def iterdir(self, **kwargs) -> Iterator[_P]:
        """
        Iterates over the directory.
        """
        uri_scheme = self._uri_scheme
        for f in self.sync_fs.glob(self._cpath_str, **kwargs):
            if self.is_cloud: yield self._new(f'{uri_scheme}://' + f)
            else: yield self._new(f)

    def get_modified(self, version_id: str = None, refresh: bool = False, **kwargs):
        """
        Return the last modified timestamp of file at path as a datetime
        """
        return self.sync_fs.modified(self._cpath_str, version_id=version_id, refresh=refresh, **kwargs)
    
    ## return self.sync_fs.modified(self._cpath_str, 

    def ls(self, detail: bool = False, **kwargs):
        """
        List objects at path.
        """
        return self.sync_fs.ls(self._cpath_str, detail=detail, **kwargs)

    
    def mkdir(self, mode: int = 0o777, parents: bool = False, exist_ok: bool = False) -> None:
        """
        Create a new directory at this given path.
        """
        ## Will return if exist_ok first, otherwise will fail check
        if self.exists() and exist_ok: return
        return self.sync_fs.mkdir(self._cpath_str, mode=mode, create_parents=parents, exist_ok=exist_ok)

    def move(self, dest: base.PathLike, recursive: bool = False, maxdepth: Optional[int] = None, overwrite: bool = False, skip_errors: bool = False, **kwargs):
        """
        Move file(s) from one location to another
        """
        _dest = self._get_dest(dest, recursive=recursive, overwrite=overwrite, skip_errors=skip_errors)
        self.sync_fs.move(self._cpath_str, _dest._cpath_str, recursive = recursive, maxdepth=maxdepth)
        return _dest

    
    def mv(self, dest: base.PathLike, recursive: bool = False, maxdepth: Optional[int] = None, overwrite: bool = False, skip_errors: bool = False, **kwargs):
        """
        Move file(s) from one location to another
        """
        _dest = self._get_dest(dest, recursive=recursive, overwrite=overwrite, skip_errors=skip_errors)
        self.sync_fs.move(self._cpath_str, _dest._cpath_str, recursive = recursive, maxdepth=maxdepth)
        return _dest

    def open(self, mode: str = 'r', encoding: Optional[str] = base.DEFAULT_ENCODING, errors: Optional[str] = None, block_size: int = 5242880, compression: str = 'infer', **kwargs: Any) -> typing.IO[Union[str, bytes]]:
        """Opens the file."""
        filelike = self.sync_fs.open(self._cpath_str, mode=mode, encoding=encoding, errors=errors, block_size=block_size, compression=compression, **kwargs)
        filelike = typing.cast(typing.IO[Union[str, bytes]], filelike)
        return filelike

    def put_file(self, dest: base.PathLike, callback: Optional[Callable] = None, overwrite: bool = False, skip_errors: bool = False, **kwargs):
        """
        Copy file from local.
        """
        _dest = self._get_dest(dest, overwrite=overwrite, skip_errors=skip_errors)
        self.sync_fs.put_file(self._cpath_str, _dest._cpath_str, callback = callback, **kwargs)
        return _dest
    
    def put_files(self, dest: base.PathLike, recursive: bool = False, callback: Optional[Callable] = None, overwrite: bool = False, skip_errors: bool = False, **kwargs):
        """
        Copy file(s) from local.
        """
        _dest = self._get_dest(dest, recursive=recursive, overwrite=overwrite, skip_errors=skip_errors)
        self.sync_fs.put(self._cpath_str, _dest._cpath_str, recursive=recursive, callback = callback, **kwargs)
        return _dest

    def remove(self, recursive: bool = False, maxdepth: Optional[int] = None, **kwargs):
        """
        Delete files.
        """
        return self.sync_fs.rm(self._cpath_str, recursive=recursive, maxdepth = maxdepth, **kwargs)

    def rename(self, dest: base.PathLike, recursive: bool = False, maxdepth: Optional[int] = None, overwrite: bool = False, skip_errors: bool = False, **kwargs):
        """
        Rename file or directory to the given target.
        """
        _dest = self._get_dest(dest, recursive=recursive, overwrite=overwrite, skip_errors=skip_errors)
        self.sync_fs.rename(self._cpath_str, _dest._cpath_str, recursive=recursive, maxdepth = maxdepth, **kwargs)
        return _dest

    def replace(self, target: base.PathLike) -> _P:
        """
        Replace file or directory to the given target.
        """
        _dest = self._get_dest(target, overwrite=True)
        self.sync_fs.rename(self._cpath_str, _dest._cpath_str)

    def rm(self, recursive: bool = False, maxdepth: Optional[int] = None, **kwargs):
        """
        Delete files.
        """
        self.sync_fs.rm(self._cpath_str, recursive, maxdepth, **kwargs)
    
    def rm_file(self, missing_ok: bool = True, **kwargs):
        """
        Delete a file
        """
        if not self.exists():
            if missing_ok: return
            raise ValueError(f"{self._path_str} does not exist")
        self.sync_fs.rm_file(self._cpath_str, **kwargs)

    def rmdir(self, force: bool = False, recursive: bool = True, skip_errors: bool = True) -> None:
        """
        Remove the empty directory.
        If force, will recursively remove even if not empty.
        """
        try:
            return self.sync_fs.rmdir(self._cpath_str)
        except Exception as e:
            if force: return self.sync_fs.rm(self._cpath_str, recursive = recursive)
            if skip_errors: return
            raise e

    def rmtree(self, recursive: bool = True, maxdepth: Optional[int] = None) -> None:
        """Remove the directory."""
        return self.sync_fs.rm(self._cpath_str, recursive, maxdepth)
    
    def sign(self, expiration: int = 100, **kwargs):
        """
        Create a signed URL representing the given path
        Some implementations allow temporary URLs to be generated, as a way of delegating credentials
        """
        return self.sync_fs.sign(self._cpath_str, expiration, **kwargs)
    
    def split_path(self, **kwargs) -> List[str]:
        """
        Normalise path string into bucket and key
        """
        return self.sync_fs.split_path(self._cpath_str, **kwargs)
    
    def tail(self, size: Optional[int] = None):
        """ 
        Get the last size bytes from file
        """
        return self.sync_fs.tail(self._cpath_str, size)

    def touch(self, truncate: bool = True, data = None, **kwargs):
        """
        Create empty file or truncate
        """
        return self.sync_fs.touch(self._cpath_str, truncate = truncate, data = data, **kwargs)
    
    def unlink(self, missing_ok: bool = True, **kwargs) -> None:
        """
        Remove this file or symbolic link.
        """
        if getattr(self.sync_fs, 'unlink', None): return self.sync_fs.unlink(self._cpath_str, missing_ok = missing_ok, **kwargs)
        return self.rm_file(missing_ok, **kwargs)
    
    def url(self, expires: int = 3600, client_method: str = 'get_object', **kwargs):
        """
        Generate presigned URL to access path by HTTP
        """
        return self.sync_fs.url(self._cpath_str, expires, client_method, **kwargs)

    def upload(self, dest: base.PathLike, recursive: bool = False, callback: Optional[Callable] = None, overwrite: bool = False, skip_errors: bool = False, **kwargs):
        """
        Copy file(s) from local.
        """
        _dest = self._get_dest(dest, recursive=recursive, overwrite=overwrite, skip_errors=skip_errors)
        return self.sync_fs.url(self._cpath_str, _dest._cpath_str, recursive=recursive, callback=callback, **kwargs)

    def get_bucket(self, prefix: bool = True) -> Optional[str]:
        """
        Returns the root bucket with optional prefix
        """
        if not self.is_cloud: return None
        parts = self.split_path()
        uri_scheme = self._uri_scheme
        if prefix: return f'{uri_scheme}://' + parts[0]
        return parts[0]
    
    def get_bucket_path(self, prefix: bool = True) -> Optional[str]:
        """
        Returns the root bucket + path with optional prefix
        removes the versioning
        """
        if not self.is_cloud: return None
        parts = self.split_path()
        uri_scheme = self._uri_scheme
        p = '/'.join(parts[:2])
        if prefix: return f'{uri_scheme}://' + p
        return p
    
    def resolve(self, strict: bool = False):
        """
        Returns the abolute path.
        """
        if self.is_cloud: return self._new(self.as_posix())
        return self._new(self._PATH.abspath(self._path_str))

    @property
    def bucket(self) -> Optional[str]:
        if not self.is_cloud: return None
        parts = self.split_path()
        return parts[0]

    @property
    def bucket_path(self) -> Optional[str]:
        if not self.is_cloud: return None
        parts = self.split_path()
        return parts[1]

    @property
    def _path_str(self) -> str:
        """
        Returns the `__fspath__` string representation.
        """
        uri_scheme = self._uri_scheme
        if uri_scheme: return self._PATH.join(f'{uri_scheme}://', *self.parts[2:])
        else: return self._PATH.join(*self.parts) if self.parts else '.'
    
    @property
    def string(self) -> str:
        """
        Returns the extension for a file
        """
        return self._path_str

    @property
    def _filename_str(self) -> str:
        """
        Returns the filename if is file, else ''
        """
        if self.is_file(): return self.parts[-1]
        return ''

    @property
    def extension(self) -> str:
        """
        Returns the extension for a file
        """
        return self.suffix

    @property
    def _cpath_str(self) -> str:
        """
        Returns the `__fspath__` string representation without the uri_scheme
        """
        uri_scheme = self._uri_scheme
        if uri_scheme: return self._PATH.join(*self.parts[2:])
        else: return self._PATH.join(*self.parts) if self.parts else '.'


    @property
    def checksum(self):
        """
        Unique value for current version of file without refreshing
        """
        return self.get_checksum()
    
    @property
    def created(self):
        """
        Return the created timestamp of a file as a datetime
        """
        return self.get_checksum()
    

    @property
    def exist(self) -> bool:
        """
        Returns True if self exists.
        """
        return self.exists()

    @property
    def info(self):
        """
        Give details of entry at path
        """
        return self.sync_fs.info(self._cpath_str)
    
    
    
    @property
    def isdir(self) -> bool:
        """
        Returns True if self is a directory.
        """
        return self.is_dir()

    @property
    def isfile(self) -> bool:
        """
        Returns True if self is a file.
        """
        return self.is_file()

    @property
    def modified(self):
        """
        Return the last modified timestamp of file at path as a datetime without refreshing
        """
        return self.get_modified()
    
    @property
    def size(self):
        """
        Size in bytes of file
        """
        return self.sync_fs.size(self._cpath_str)
    
    @property
    def stat(self):
        """
        Give details of entry at path
        """
        return self.sync_fs.stat(self._cpath_str)
    
    @property
    def home(self) -> Optional[str]:
        """
        Returns the home directory/bucket
        """
        if self.is_cloud: return self._new(self.get_bucket(True))
        p = os.path.expanduser('~')
        return self._new(p)
    
    ###############################################################
    ####         Async Versions of the Sync Functions           ###
    ###############################################################

    async def async_info(self):
        """
        Give details of entry at path
        """
        return await self.async_fs._info(self._cpath_str)

    async def async_exists(self) -> bool:
        """
        Returns True if self exists.
        """
        return await self.async_fs._exists(self._cpath_str)

    async def async_is_dir(self) -> bool:
        """
        Returns True if self is a directory.
        """
        return await self.async_fs._isdir(self._cpath_str)

    async def async_is_file(self) -> bool:
        """Returns True if self is a file."""
        return await self.async_fs._isfile(self._cpath_str)


    async def async_cat(self, recursive: bool = False, on_error: Optional[str] = None, **kwargs):
        """
        Fetch (potentially multiple) paths contents
        """
        return await self.async_fs._cat(self._cpath_str, recursive, on_error, **kwargs)
    
    async def async_cat_file(self, start: Optional[int] = None, end: Optional[int] = None, **kwargs):
        """
        Get the content of a file
        """
        return await self.async_fs._cat_file(self._cpath_str, start=start, end=end, **kwargs)

    async def async_copy(self, dest: base.PathLike, recursive: bool = False, overwrite: bool = False, skip_errors: bool = False) -> _P:
        """
        Copies the File to the Dir/File.
        """
        _dest = self._get_dest(dest, recursive, overwrite, skip_errors)
        await self.async_fs._copy(self._cpath_str, _dest._cpath_str, recursive = recursive)
        return _dest

    
    async def async_download(self, dest: base.PathLike, recursive: bool = False, callback: Optional[Callable] = None, overwrite: bool = False, skip_errors: bool = False, **kwargs):
        """
        Copy file(s) to local.
        """
        _dest = self._get_dest(dest, recursive, overwrite, skip_errors)
        await self.async_fs._download(self._cpath_str, _dest._cpath_str, recursive = recursive, callback = callback)
        return _dest
    
    async def async_du(self, total: Optional[int] = None, maxdepth: Optional[int] = None, **kwargs):
        """
        Fetch (potentially multiple) paths contents
        """
        return await self.async_fs._du(self._cpath_str, total, maxdepth, **kwargs)
    
    async def async_expand_path(self, recursive: bool = False, maxdepth: Optional[int] = None):
        """
        Turn one or more globs or directories into a list of all matching paths to files or directories.
        """
        paths = await self.async_fs._expand_path(recursive = recursive, maxdepth = maxdepth)
        return [self._new(p) for p in paths]
    
    async def async_find(self, path: Optional[str] = None, maxdepth: Optional[int] = None, withdirs: bool = False, detail: bool = False, prefix: Optional[str] = None, **kwargs):
        """
        List all files below path.
        """
        p = self if not path else self.joinpath(path)
        return await self.async_fs._find(p, maxdepth=maxdepth, withdirs=withdirs, detail=detail, prefix=prefix, **kwargs)
    
    async def async_get_checksum(self, refresh: bool = False, **kwargs):
        """
        Unique value for current version of file
        """
        return await self.async_fs.checksum(self._cpath_str, **kwargs)
        #return self.async_fs.checksum(self._cpath_str, refresh=refresh, **kwargs)
    
    async def async_get_file(self, dest: base.PathLike, callback: Optional[Callable] = None, overwrite: bool = False, skip_errors: bool = False, **kwargs):
        """
        Copy single remote file to local
        """
        _dest = self._get_dest(dest, overwrite=overwrite, skip_errors=skip_errors)
        return await self.async_fs._get_file(self._cpath_str, _dest._cpath_str, callback=callback, **kwargs)


    async def async_get_files(self, dest: base.PathLike, recursive: bool = False, callback: Optional[Callable] = None, overwrite: bool = False, skip_errors: bool = False, **kwargs):
        """
        Copy file(s) to local.
        """
        _dest = self._get_dest(dest, recursive=recursive, overwrite=overwrite, skip_errors=skip_errors)
        return await self.async_fs._get(self._cpath_str, _dest._cpath_str, recursive=recursive, callback=callback, **kwargs)
        

    async def async_glob(self, pattern: str, **kwargs) -> Iterator[_P]:
        """
        Yielding all matching files (of any kind).
        """
        uri_scheme = self._uri_scheme
        for f in await self.async_fs._glob(self._PATH.join(self._path_str, pattern)._cpath_str,  **kwargs):
            if self.is_cloud: yield self._new(f'{uri_scheme}://' + f)
            else: yield self._new(f)

    async def async_head(self, size: Optional[int] = None):
        """ 
        Get the first size bytes from file
        """
        return self.async_fs.head(self._cpath_str, size)

    async def async_iterdir(self, **kwargs) -> Iterator[_P]:
        """
        Iterates over the directory.
        """
        uri_scheme = self._uri_scheme
        for f in await self.async_fs._glob(self._cpath_str, **kwargs):
            if self.is_cloud: yield self._new(f'{uri_scheme}://' + f)
            else: yield self._new(f)

    async def async_get_modified(self, version_id: str = None, refresh: bool = False, **kwargs):
        """
        Return the last modified timestamp of file at path as a datetime
        """
        return self.async_fs.modified(self._cpath_str, version_id=version_id, refresh=refresh, **kwargs)

    async def async_ls(self, detail: bool = False, **kwargs):
        """
        List objects at path.
        """
        return await self.async_fs._ls(self._cpath_str, detail=detail, **kwargs)

    
    async def async_mkdir(self, mode: int = 0o777, parents: bool = False, exist_ok: bool = False) -> None:
        """
        Create a new directory at this given path.
        """
        if self.exists() and exist_ok: return
        return await self.async_fs._mkdir(self._cpath_str, mode=mode, create_parents=parents, exist_ok=exist_ok)

    async def async_move(self, dest: base.PathLike, recursive: bool = False, maxdepth: Optional[int] = None, overwrite: bool = False, skip_errors: bool = False, **kwargs):
        """
        Move file(s) from one location to another
        """
        _dest = self._get_dest(dest, recursive=recursive, overwrite=overwrite, skip_errors=skip_errors)
        self.async_fs.move(self._cpath_str, _dest._cpath_str, recursive = recursive, maxdepth=maxdepth)
        return _dest

    
    async def async_mv(self, dest: base.PathLike, recursive: bool = False, maxdepth: Optional[int] = None, overwrite: bool = False, skip_errors: bool = False, **kwargs):
        """
        Move file(s) from one location to another
        """
        _dest = self._get_dest(dest, recursive=recursive, overwrite=overwrite, skip_errors=skip_errors)
        self.async_fs.move(self._cpath_str, _dest._cpath_str, recursive = recursive, maxdepth=maxdepth)
        return _dest

    async def async_open(self, mode: str = 'r', encoding: Optional[str] = base.DEFAULT_ENCODING, errors: Optional[str] = None, block_size: int = 5242880, compression: str = 'infer', **kwargs: Any) -> typing.IO[Union[str, bytes]]:
        """Opens the file."""
        filelike = await self.async_fs._open(self._cpath_str, mode=mode, encoding=encoding, errors=errors, block_size=block_size, compression=compression, **kwargs)
        filelike = typing.cast(typing.IO[Union[str, bytes]], filelike)
        return filelike

    async def async_put_file(self, dest: base.PathLike, callback: Optional[Callable] = None, overwrite: bool = False, skip_errors: bool = False, **kwargs):
        """
        Copy file from local.
        """
        _dest = self._get_dest(dest, overwrite=overwrite, skip_errors=skip_errors)
        await self.async_fs._put_file(self._cpath_str, _dest._cpath_str, callback = callback, **kwargs)
        return _dest
    
    async def async_put_files(self, dest: base.PathLike, recursive: bool = False, callback: Optional[Callable] = None, overwrite: bool = False, skip_errors: bool = False, **kwargs):
        """
        Copy file(s) from local.
        """
        _dest = self._get_dest(dest, recursive=recursive, overwrite=overwrite, skip_errors=skip_errors)
        await self.async_fs._put(self._cpath_str, _dest._cpath_str, recursive=recursive, callback = callback, **kwargs)
        return _dest

    async def async_remove(self, recursive: bool = False, maxdepth: Optional[int] = None, **kwargs):
        """
        Delete files.
        """
        return await self.async_fs._rm(self._cpath_str, recursive=recursive, maxdepth = maxdepth, **kwargs)

    async def async_rename(self, dest: base.PathLike, recursive: bool = False, maxdepth: Optional[int] = None, overwrite: bool = False, skip_errors: bool = False, **kwargs):
        """
        Rename file or directory to the given target.
        """
        _dest = self._get_dest(dest, recursive=recursive, overwrite=overwrite, skip_errors=skip_errors)
        self.async_fs.rename(self._cpath_str, _dest._cpath_str, recursive=recursive, maxdepth = maxdepth, **kwargs)
        return _dest

    async def async_replace(self, target: base.PathLike) -> _P:
        """
        Replace file or directory to the given target.
        """
        _dest = self._get_dest(target, overwrite=True)
        self.async_fs.rename(self._cpath_str, _dest._cpath_str)

    async def async_rm(self, recursive: bool = False, maxdepth: Optional[int] = None, **kwargs):
        """
        Delete files.
        """
        await self.async_fs._rm(self._cpath_str, recursive, maxdepth, **kwargs)
    
    async def async_rm_file(self, missing_ok: bool = True, **kwargs):
        """
        Delete a file
        """
        if not self.exists():
            if missing_ok: return
            raise ValueError(f"{self._path_str} does not exist")
        await self.async_fs._rm_file(self._cpath_str, **kwargs)

    async def async_rmdir(self, force: bool = False, recursive: bool = True, skip_errors: bool = True) -> None:
        """
        Remove the empty directory.
        If force, will recursively remove even if not empty.
        """
        try:
            return self.async_fs.rmdir(self._cpath_str)
        except Exception as e:
            if force: return await self.async_fs._rm(self._cpath_str, recursive = recursive)
            if skip_errors: return
            raise e

    async def async_rmtree(self, recursive: bool = True, maxdepth: Optional[int] = None) -> None:
        """Remove the directory."""
        return await self.async_fs._rm(self._cpath_str, recursive, maxdepth)
    
    async def async_sign(self, expiration: int = 100, **kwargs):
        """
        Create a signed URL representing the given path
        Some implementations allow temporary URLs to be generated, as a way of delegating credentials
        """
        return self.async_fs.sign(self._cpath_str, expiration, **kwargs)
    
    async def async_size(self, **kwargs):
        """
        Create a signed URL representing the given path
        Some implementations allow temporary URLs to be generated, as a way of delegating credentials
        """
        return await self.async_fs._size(self._cpath_str, **kwargs)
    
    async def async_split_path(self, **kwargs) -> List[str]:
        """
        Normalise path string into bucket and key
        """
        return self.async_fs.split_path(self._cpath_str, **kwargs)
    
    async def async_tail(self, size: Optional[int] = None):
        """ 
        Get the last size bytes from file
        """
        return self.async_fs.tail(self._cpath_str, size)

    async def async_touch(self, truncate: bool = True, data = None, **kwargs):
        """
        Create empty file or truncate
        """
        return await self.async_fs.touch(self._cpath_str, truncate = truncate, data = data, **kwargs)
    
    async def async_unlink(self, missing_ok: bool = True, **kwargs) -> None:
        """
        Remove this file or symbolic link.
        """
        if getattr(self.async_fs, 'unlink', None): return self.async_fs.unlink(self._cpath_str, missing_ok = missing_ok, **kwargs)
        return await self.async_rm_file(missing_ok, **kwargs)
    
    async def async_url(self, expires: int = 3600, client_method: str = 'get_object', **kwargs):
        """
        Generate presigned URL to access path by HTTP
        """
        return self.async_fs.url(self._cpath_str, expires, client_method, **kwargs)

    async def async_upload(self, dest: base.PathLike, recursive: bool = False, callback: Optional[Callable] = None, overwrite: bool = False, skip_errors: bool = False, **kwargs):
        """
        Copy file(s) from local.
        """
        _dest = self._get_dest(dest, recursive=recursive, overwrite=overwrite, skip_errors=skip_errors)
        return self.async_fs.url(self._cpath_str, _dest._cpath_str, recursive=recursive, callback=callback, **kwargs)


class WindowsFSxPath(PosixFSxPath, pathlib.PureWindowsPath):
    _PATH = ntpath


class PosixIOPath(PosixFSxPath, pathlib.PurePosixPath):
    """
    Pathlib-like API around `upath` providing compatability with many i/o. Does not provide Async
    """
    _PATH = posixpath
    _FSX: 'upath.UPath' = None
    _SYNC_FS: 'upath.UPath' = None
    _ASYNC_FS: Any = None
    _FSX_LIB: str = 'upath'
    _FSX_MODULE: Optional[str] = None
    _FSX_CLS: str = 'UPath'


    @property
    def async_fs(self) -> 'AsyncFileSystem':
        raise NotImplementedError
    
    @property
    def sync_fs(self) -> 'upath.UPath':
        return upath.UPath


class WindowsIOPath(PosixIOPath, pathlib.PureWindowsPath):
    """
    Pathlib-like API around `upath` providing compatability with many i/o. Does not provide Async
    """
    _PATH = ntpath

    

os.PathLike.register(PosixFSxPath)
os.PathLike.register(WindowsFSxPath)

os.PathLike.register(PosixIOPath)
os.PathLike.register(WindowsIOPath)




"""
Wrapper class around pathlib.Path
with enhancements.
This takes several of the functions found in
lazycls.utils and builds it directly into Path.

Will deprecate features in .utils in a few versions.
"""
import os
import aiofiles
import tempfile
from pathlib import Path as _Path
from typing import TypeVar, Union, Type, List, Any, Optional, Callable
from lazycls.prop import classproperty
from lazycls.serializers import Json, OrJson, Yaml, Pkl, Base

from errno import EINVAL, ENOENT, ENOTDIR, EBADF, ELOOP
from stat import S_ISDIR, S_ISLNK, S_ISREG, S_ISSOCK, S_ISBLK, S_ISCHR, S_ISFIFO

try: import httpx as _req
except ImportError: _req = None

if _req is None:
    try: import requests as _req
    except ImportError: _req = None



_IGNORED_ERROS = (ENOENT, ENOTDIR, EBADF, ELOOP)
_IGNORED_WINERRORS = (
    21,  # ERROR_NOT_READY - drive exists but is not accessible
    123, # ERROR_INVALID_NAME - fix for bpo-35306
    1921,  # ERROR_CANT_RESOLVE_FILENAME - fix for broken symlink pointing to itself
)

def _ignore_error(exception):
    return (getattr(exception, 'errno', None) in _IGNORED_ERROS or getattr(exception, 'winerror', None) in _IGNORED_WINERRORS)

_LOADERS = {
    '.json': OrJson,
    '.yml': Yaml,
    '.yaml': Yaml,
    '.pkl': Pkl,
    '.pickle': Pkl,
    '.pb': Pkl,
    'json': OrJson,
    'yaml': Yaml,
    'yml': Yaml,
    'pkl': Pkl,
    'pickle': Pkl
}

# https://stackoverflow.com/questions/29850801/subclass-pathlib-path-fails

class Path(_Path):
    _flavour = type(_Path())._flavour

    def __new__(cls, *args):
        return super(Path, cls).__new__(cls, *args)

    def __init__(self, *args):
        super().__init__()
    
    @classproperty
    def async_os(cls):
        """ Aliases for aiofiles.os"""
        return aiofiles.os
        
    @classproperty
    def working_dir(cls) -> Type[_Path]:
        return cls(os.getcwd())
    
    @classproperty
    def userhome(cls) -> Type[_Path]:
        """Return a new path pointing to the user's home directory (as
        returned by os.path.expanduser('~')).
        """
        return cls(cls()._flavour.gethomedir(None))

    @staticmethod
    def get_path(path: Union[str, _Path], resolve: bool = True, ensure_exists: bool = False, mode=0o777, file_mode=0o666, parents=False, exist_ok: bool = True) -> Type[_Path]:
        if isinstance(path, str): path = Path(path)
        if resolve: path.resolve()
        if ensure_exists and path.is_dir(): path.mkdir(mode=mode, parents=parents, exist_ok=exist_ok)
        elif ensure_exists: path.touch(mode=file_mode, exist_ok=exist_ok)
        return path

    @staticmethod
    def get_user_path(path: Union[str, _Path], resolve: bool = False) -> Type[_Path]:
        if isinstance(path, str): path = Path(path)
        path = path.expanduser()
        if resolve: path.resolve()
        return path
    
    @classmethod
    def from_env(cls, key: str, default: Union[str, _Path], **kwargs) -> Type[_Path]:
        """ gets the path from env value, returns default if none"""
        val = os.getenv(key, None)
        if val is None: return cls.get_path(default, **kwargs)
        return cls.get_path(val, **kwargs)
    
    @property
    def _exists(self): 
        return self.exists()

    @classmethod
    def from_url(cls, url: str, path: Union[str, _Path], overwrite: bool = False, **kwargs) -> Type[_Path]:
        """ Downloads a file from url and saves to  file """
        p = cls.get_path(path, resolve=True)
        if p.exists() and not overwrite: return p
        assert _req is not None, 'Requests and/or httpx is not available'
        r = _req.get(url=url, **kwargs)
        if r.status_code >= 400: raise Exception(f'url = {url} returned status_code = {r.status_code}')
        p.write_bytes(data=r.content)
        return p
    
    @staticmethod
    def get_parent_path(path: Union[str, _Path]) -> Type[_Path]:
        p = Path.get_path(path)
        return p.parent

    @staticmethod
    def get_cwd(*paths, posix: bool = True) -> Union[str, Type[_Path]]:
        if not paths:
            if posix: return Path.cwd.string
            return Path.cwd
        if posix: return Path.cwd.joinpath(*paths).string
        return Path.cwd.joinpath(*paths)

    @property
    def string(self) -> str:
        return self.as_posix()

    @property
    def extension(self):
        """ alias for .suffix or the file extension"""
        return self.suffix
    

    def reader(self, mode: str = 'r', encoding: str = 'utf-8', errors: str = None, buffering: int = -1, newline: str = None):
        """ Alias for open(r)"""
        return self.open(mode=mode, encoding=encoding, errors=errors, buffering=buffering, newline=newline)

    def binary_reader(self, mode: str = 'rb', encoding: str = 'utf-8', errors: str = None, buffering: int = -1, newline: str = None):
        """ Alias for open(rb)"""
        return self.open(mode=mode, encoding=encoding, errors=errors, buffering=buffering, newline=newline)
    
    def appender(self, mode: str = 'a', encoding: str = 'utf-8', errors: str = None, buffering: int = -1, newline: str = None):
        """ Alias for open(a)"""
        return self.open(mode=mode, encoding=encoding, errors=errors, buffering=buffering, newline=newline)
    
    def binary_appender(self, mode: str = 'ab', encoding: str = 'utf-8', errors: str = None, buffering: int = -1, newline: str = None):
        """ Alias for open(ab)"""
        return self.open(mode=mode, encoding=encoding, errors=errors, buffering=buffering, newline=newline)
    
    def writer(self, mode: str = 'w', encoding: str = 'utf-8', errors: str = None, buffering: int = -1, newline: str = None):
        """ Alias for open(w)"""
        return self.open(mode=mode, encoding=encoding, errors=errors, buffering=buffering, newline=newline)
    
    def binary_writer(self, mode: str = 'wb', encoding: str = 'utf-8', errors: str = None, buffering: int = -1, newline: str = None):
        """ Alias for open(wb)"""
        return self.open(mode=mode, encoding=encoding, errors=errors, buffering=buffering, newline=newline)


    async def async_open(self, mode: str = 'r', encoding: str = 'utf-8', errors: str = None, buffering: int = -1, newline: str = None, closefd: bool = True, opener: Optional[Callable[[str, int], int]] = None, **kwargs):
        """ Alias for aiofiles.open(r)"""
        return await aiofiles.open(self.string, mode=mode, encoding=encoding, errors=errors, buffering=buffering, newline=newline, closefd=closefd, opener=opener, **kwargs)
    
    async def async_reader(self, mode: str = 'r', encoding: str = 'utf-8', errors: str = None, buffering: int = -1, newline: str = None, closefd: bool = True, opener: Optional[Callable[[str, int], int]] = None, **kwargs):
        """ Alias for aiofiles.open(r)"""
        return await aiofiles.open(self.string, mode=mode, encoding=encoding, errors=errors, buffering=buffering, newline=newline, closefd=closefd, opener=opener, **kwargs)
    
    async def async_reader_binary(self, mode: str = 'rb', encoding: str = 'utf-8', errors: str = None, buffering: int = -1, newline: str = None, closefd: bool = True, opener: Optional[Callable[[str, int], int]] = None, **kwargs):
        """ Alias for aiofiles.open(rb)"""
        return await aiofiles.open(self.string, mode=mode, encoding=encoding, errors=errors, buffering=buffering, newline=newline, closefd=closefd, opener=opener, **kwargs)
    
    async def async_appender(self, mode: str = 'a', encoding: str = 'utf-8', errors: str = None, buffering: int = -1, newline: str = None, closefd: bool = True, opener: Optional[Callable[[str, int], int]] = None, **kwargs):
        """ Alias for aiofiles.open(a)"""
        return await aiofiles.open(self.string, mode=mode, encoding=encoding, errors=errors, buffering=buffering, newline=newline, closefd=closefd, opener=opener, **kwargs)
    
    async def async_appender_binary(self, mode: str = 'ab', encoding: str = 'utf-8', errors: str = None, buffering: int = -1, newline: str = None, closefd: bool = True, opener: Optional[Callable[[str, int], int]] = None, **kwargs):
        """ Alias for aiofiles.open(ab)"""
        return await aiofiles.open(self.string, mode=mode, encoding=encoding, errors=errors, buffering=buffering, newline=newline, closefd=closefd, opener=opener, **kwargs)
    
    async def async_writer(self, mode: str = 'w', encoding: str = 'utf-8', errors: str = None, buffering: int = -1, newline: str = None, closefd: bool = True, opener: Optional[Callable[[str, int], int]] = None, **kwargs):
        """ Alias for aiofiles.open(w)"""
        return await aiofiles.open(self.string, mode=mode, encoding=encoding, errors=errors, buffering=buffering, newline=newline, closefd=closefd, opener=opener, **kwargs)
    
    async def async_writer_binary(self, mode: str = 'wb', encoding: str = 'utf-8', errors: str = None, buffering: int = -1, newline: str = None, closefd: bool = True, opener: Optional[Callable[[str, int], int]] = None, **kwargs):
        """ Alias for aiofiles.open(wb)"""
        return await aiofiles.open(self.string, mode=mode, encoding=encoding, errors=errors, buffering=buffering, newline=newline, closefd=closefd, opener=opener, **kwargs)
    
    @staticmethod
    def get_tempfile(mode: str = 'w', encoding: str = 'utf-8', dir: Any = None, buffering: int = -1, newline: str = None, delete: bool = False, **kwargs):
        return tempfile.NamedTemporaryFile(mode=mode, encoding=encoding, buffering=buffering, newline=newline, delete=delete, dir=dir, **kwargs)

    @staticmethod
    def get_tempdir(dir: Any = None, **kwargs):
        return tempfile.TemporaryDirectory(dir=dir, **kwargs)

    @staticmethod
    async def async_get_tempfile(mode: str = 'w', encoding: str = 'utf-8', dir: Any = None, buffering: int = -1, newline: str = None, delete: bool = False, **kwargs):
        return await aiofiles.tempfile.NamedTemporaryFile(mode=mode, encoding=encoding, buffering=buffering, newline=newline, delete=delete, dir=dir, **kwargs)

    @staticmethod
    async def async_get_tempdir(dir: Any = None, **kwargs):
        return await aiofiles.tempfile.TemporaryDirectory(dir=dir, **kwargs)

    def to_base64(self, **kwargs):
        """ Reads the file as text, then turns to base64"""
        data = self.read_text(**kwargs)
        return Base.b64_encode(data)
    
    def to_gzip_base64(self, **kwargs):
        """ Reads the file as text, then turns to gzip+base64"""
        data = self.read_text(**kwargs)
        return Base.b64_gzip_encode(data)
        
    @classmethod
    def from_base64(cls, data: Union[str, bytes], path: Union[str, _Path] = None, encoding: str = 'utf-8', errors: str = None, **kwargs):
        d = Base.b64_decode(data)
        if not path: return d
        p = cls.get_path(path, **kwargs)
        p.write_text(data=d, encoding=encoding, errors=errors)
        return p

    @classmethod
    def from_gzip_base64(cls, data: Union[str, bytes], path: Union[str, _Path] = None, encoding: str = 'utf-8', errors: str = None, **kwargs):
        d = Base.b64_gzip_decode(data)
        if not path: return d
        p = cls.get_path(path, **kwargs)
        p.write_text(data=d, encoding=encoding, errors=errors)
        return p

    def read_bytes(self):
        """
        Open the file in bytes mode (rb), read it, and close the file.
        """
        with self.open(mode='rb') as f:
            return f.read()

    def read_text(self, encoding: str = 'utf-8', errors: str = None):
        """
        Open the file in text mode (r), read it, and close the file.
        """
        with self.open(mode='r', encoding=encoding, errors=errors) as f:
            return f.read()

    def append_bytes(self, data):
        """
        Open the file in bytes mode (ab), append to it, and close the file.
        """
        # type-check for the buffer interface before truncating the file
        view = memoryview(data)
        with self.open(mode='ab') as f:
            return f.write(view)

    def append_text(self, data: str, encoding: str ='utf-8', errors: str = None, insert_newline: bool = True):
        """
        Open the file in text mode (a), write to it, and close the file.
        """
        if not isinstance(data, str): raise TypeError('data must be str, not %s' % data.__class__.__name__)
        with self.open(mode='a', encoding=encoding, errors=errors) as f:
            f.write(data)
            if insert_newline: f.write('\n')
            return f

    def write_bytes(self, data):
        """
        Open the file in bytes mode (wb), write to it, and close the file.
        """
        # type-check for the buffer interface before truncating the file
        view = memoryview(data)
        with self.open(mode='wb') as f:
            return f.write(view)

    def write_text(self, data: str, encoding: str ='utf-8', errors: str = None):
        """
        Open the file in text mode (w), write to it, and close the file.
        """
        if not isinstance(data, str): raise TypeError('data must be str, not %s' % data.__class__.__name__)
        with self.open(mode='w', encoding=encoding, errors=errors) as f:
            return f.write(data)
    
    async def async_read_bytes(self, **kwargs):
        """
        Open the file in bytes mode (rb), read it, and close the file.
        """
        async with self.async_open(mode='rb', **kwargs) as f:
            return await f.read()

    async def async_read_text(self, encoding: str = 'utf-8', errors: str = None, **kwargs):
        """
        Open the file in text mode (r), read it, and close the file.
        """
        async with self.async_open(mode='r', encoding=encoding, errors=errors, **kwargs) as f:
            return await f.read()

    def get_files(self, pattern: str = '*') -> List[Type[_Path]]:
        """
        Uses the .glob(pattern) to construct List[Type[Path]] instead of returning an iterator
        """
        p_iter = self.glob(pattern)
        return [p for p in p_iter if p]

    def get_loader(self, loader: str = None):
        """ If the file is supported (pkl, yaml, json) - then return the loader"""
        if loader: return _LOADERS.get(loader, _LOADERS.get(self.extension, None))
        return _LOADERS.get(self.extension, None)

    def loads(self, encoding: str ='utf-8', errors: str = None, loader: str = None, binary: bool = False, **kwargs) -> Any:
        """ If the file is supported (pkl, yaml, json) - then return the loader.loads(**kwargs) method.
            else:
                - if binary: returns .read_bytes()
                    else: returns .read_text()
        """
        l = self.get_loader(loader=loader)
        if l: return l.loads(self.read_text(encoding=encoding, errors=errors), **kwargs)
        if binary: return self.read_bytes()
        return self.read_text(encoding=encoding, errors=errors)

    def dumps(self, data: Any, encoding: str ='utf-8', errors: str = None, loader: str = None, **kwargs) -> Any:
        """ If the file is supported (pkl, yaml, json) - then perform the loader.dumps(data, **kwargs) method.
            else:
                - if binary: performs .write_bytes(data)
                    else: performs .write_text(data)
            returns: Path
        """
        l = self.get_loader(loader=loader)
        if l:
            d = l.dumps(data, **kwargs)
            if isinstance(d, bytes): self.write_bytes(d)
            else: self.write_text(d, encoding=encoding, errors=errors)
        elif isinstance(data, bytes): self.write_bytes(data)
        else: self.write_text(data, encoding=encoding, errors=errors)
        return self


    """ Modifying the below to be properties rather than callables"""

    @property
    def _is_dir(self):
        """
        Whether this path is a directory.
        """
        try: return S_ISDIR(self.stat().st_mode)
        except OSError as e:
            if not _ignore_error(e): raise
            return False
        except ValueError: return False

    @property
    def _is_file(self):
        """
        Whether this path is a regular file (also True for symlinks pointing
        to regular files).
        """
        try: return S_ISREG(self.stat().st_mode)
        except OSError as e:
            if not _ignore_error(e): raise
            return False
        except ValueError: return False

    @property
    def _is_mount(self):
        """
        Check if this path is a POSIX mount point
        """
        # Need to exist and be a dir
        if not self.exists() or not self.is_dir(): return False
        try: parent_dev = self.parent.stat().st_dev
        except OSError: return False
        dev = self.stat().st_dev
        if dev != parent_dev: return True
        ino = self.stat().st_ino
        parent_ino = self.parent.stat().st_ino
        return ino == parent_ino

    @property
    def _is_symlink(self):
        """
        Whether this path is a symbolic link.
        """
        try: return S_ISLNK(self.lstat().st_mode)
        except OSError as e:
            if not _ignore_error(e): raise
            return False
        except ValueError: return False

    @property
    def _is_block_device(self):
        """
        Whether this path is a block device.
        """
        try: return S_ISBLK(self.stat().st_mode)
        except OSError as e:
            if not _ignore_error(e): raise
            return False
        except ValueError: return False

    @property
    def _is_char_device(self):
        """
        Whether this path is a character device.
        """
        try: return S_ISCHR(self.stat().st_mode)
        except OSError as e:
            if not _ignore_error(e): raise
            return False
        except ValueError: return False

    @property
    def _is_fifo(self):
        """
        Whether this path is a FIFO.
        """
        try: return S_ISFIFO(self.stat().st_mode)
        except OSError as e:
            if not _ignore_error(e): raise
            return False
        except ValueError: return False

    @property
    def _is_socket(self):
        """
        Whether this path is a socket.
        """
        try: return S_ISSOCK(self.stat().st_mode)
        except OSError as e:
            if not _ignore_error(e): raise
            return False
        except ValueError: return False


PathLike = TypeVar('PathLike', type(_Path), Path, _Path, os.PathLike, str, Union[str, _Path], Union[str, Path], Union[str, os.PathLike])

__all__ = [
    'Path',
    'PathLike'
]




"""
Cloud Provider Configs
"""
from typing import Optional
from lazy.types import pyd
from .core import ConfigCls
from .types import *


## not sure if needed
from logz import get_logger
logger = get_logger('cloudauthz')

__all__ = ('CloudAuthz')

class CloudAuthz(ConfigCls):
    authz_dir: PathStr = "./authz"
    boto_config: Optional[PathStr] = "~/.boto"
    
    """ 
    AWS Specific 
    """
    aws_access_token: Optional[str] = ""
    aws_access_key_id: Optional[str] = ""
    aws_secret_access_key: Optional[str] = ""
    aws_region: Optional[str] = "us-east-1"
    set_s3_endpoint: Optional[bool] = True
    s3_config: Optional[pyd.Json] = None

    """ 
    GCP Specific 
    """
    
    gcloud_project: str = ""
    google_cloud_project: str = ""
    gauth: GoogleAuthBGZ = "" # Kept for compatability
    gcp_auth: GoogleAuthJsonStr = ""
    gcp_authb64: GoogleAuthB64 = ""
    gcp_authbgz: GoogleAuthBGZ = ""

    gcs_client_config: Optional[pyd.Json] = None
    gcs_config: Optional[pyd.Json] = None

    """
    Minio Specific
    """
    minio_endpoint: Optional[str] = ""
    minio_access_key: Optional[str] = ""
    minio_secret_key: Optional[str] = ""
    minio_access_token: Optional[str] = ""
    minio_config: Optional[pyd.Json] = None

    """
    S3-Compatiable Generic
    """
    s3compat_endpoint: Optional[str] = ""
    s3compat_region: Optional[str] = ""
    s3compat_access_key: Optional[str] = ""
    s3compat_secret_key: Optional[str] = ""
    s3compat_access_token: Optional[str] = ""
    s3compat_config: Optional[pyd.Json] = None

    def get_s3_endpoint(self):
        return f'https://s3.{self.aws_region}.amazonaws.com'

    def get_boto_path(self):
        if self.boto_config and self.boto_config.exists(): return self.boto_config
        if not self.authz_dir.exist: self.authz_dir.mkdir(create_parents=True, exist_ok=True)
        return self.authz_dir.joinpath('.boto')
    
    def should_write_boto(self):
        return not bool(self.boto_config and self.boto_config.exists())
    
    def get_boto_values(self):
        t = "[Credentials]\n"
        if self.aws_access_key_id:
            t += f"aws_access_key_id = {self.aws_access_key_id}\n"
            t += f"aws_secret_access_key = {self.aws_secret_access_key}\n"
        if self.gauth and self.gauth.to_env_path.exists():
            t += f"gs_service_key_file = {self.gauth}\n"
        elif self.gcp_auth and self.gcp_auth.to_env_path.exists():
            t += f"gs_service_key_file = {self.gcp_auth}\n"
        t += "\n[Boto]\n"
        t += "https_validate_certificates = True\n"
        t += "\n[GSUtil]\n"
        t += "content_language = en\n"
        t += "default_api_version = 2\n"
        if self.gcloud_project or self.google_cloud_project:
            t += f"default_project_id = {self.gcloud_project or self.google_cloud_project}\n"
        return t


    def write_botofile(self, overwrite: bool = False, **kwargs):
        if self.should_write_boto:
            p = self.get_boto_path()
            if not p.exists() or overwrite:
                logger.info(f'Writing Botofile to {p.as_posix()}')
                p.write_text(self.get_boto_values())
            else: 
                logger.error(f'Botofile {p.as_posix()} exists and overwrite is False. Not overwriting')
            return p
        else: 
            logger.warning(f'Skipping writing Botofile as BotoConfig = {self.boto_config.as_posix()} exists')
            return self.boto_config


    def set_authz_env(self):
        from lazy.cmd.contrib import export
        if self.gcp_auth: export(GOOGLE_APPLICATION_CREDENTIALS=self.gcp_auth)
        elif self.gauth: export(GOOGLE_APPLICATION_CREDENTIALS=self.gauth)

        if self.gcloud_project: export(GOOGLE_CLOUD_PROJECT=self.gcloud_project or self.google_cloud_project)
        botopath = self.get_boto_path()
        ## We know this is our custom botofile
        if botopath.exists() and self.should_write_boto():
            export(BOTO_PATH=botopath.as_posix())
            export(BOTO_CONFIG=botopath.as_posix())
        if self.aws_access_key_id:
            export(AWS_ACCESS_KEY_ID=self.aws_access_key_id)
            export(AWS_SECRET_ACCESS_KEY=self.aws_secret_access_key)
        if self.set_s3_endpoint:
            export(S3_ENDPOINT=self.get_s3_endpoint())
        if self.minio_access_key:
            export(MINIO_ACCESS_KEY=self.minio_access_key)
            export(MINIO_SECRET_KEY=self.minio_secret_key)
        if self.minio_endpoint:
            export(MINIO_ENDPOINT=self.minio_endpoint)

    def update_authz(self, **config):
        self.update_config(**config)
        self.set_authz_env()
        
        


        


"""
Built using https://github.com/kbairak/pipepy as base/inspiration and extends library

Credits to https://github.com/kbairak

Contrib extensions include applying async operations using AnyIO
"""

from .core import Cmd
from . import auto
from .contrib import cd, export, source
